/*
 * drivers/mmc/card/aerial/aerial_drv.c
 *
 *  Copyright (C) 2008 Nissin Systems Co.,Ltd.
 *  Copyright (C) 2008-2009 Atmark Techno, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 2008-03-05    Created by Nissin Systems Co.,Ltd.
 * 2009-06-19   Modified by Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/kthread.h>
#include <linux/etherdevice.h>
#include <linux/leds.h>
#include <linux/mmc/host.h>
#include <linux/mmc/card.h>
#include <linux/mmc/sdio.h>
#include <linux/mmc/sdio_func.h>
#include <linux/mmc/sdio_ids.h>

#include "aerial_drv.h"
#include "aerial_wid.h"
#include "aerial_ioctl.h"
#include "aerial_fw.h"
#include "aerial_sysfs.h"

#define IS_TXWAIT(v) \
	(((v) & AERIAL_CISSTAT_MASK) == AERIAL_CISSTAT_TXWAIT)
#define IS_TXBUSY(v) \
	((((v) & AERIAL_CISSTAT_MASK) == AERIAL_CISSTAT_TXSLEEP) || \
	 (((v) & AERIAL_CISSTAT_MASK) == AERIAL_CISSTAT_TXACTIVE))
#define IS_RXRDY(v) \
	(((v) & AERIAL_CISSTAT_MASK) == AERIAL_CISSTAT_RXREADY)
#define IS_RXNODATA(v) \
	(((v) & AERIAL_CISSTAT_MASK) == AERIAL_CISSTAT_RXNODATA)
#define IS_TOGGLE(v,t) \
	(((t) == TOGGLE_NONE) ? 1 : ((((v) >> AERIAL_TOGGLE_SHIFT) & 0x1) == t))
#define SET_TOGGLE(v,t) \
	{ (t) = (((v) >> AERIAL_TOGGLE_SHIFT) & 0x01) ^ 0x01; }

struct sdio_bus_type {
	int id;
#define SDIO_DYNAMIC_WIDTH	(0)
#define SDIO_FORCE_1BIT		(1)
#define SDIO_FORCE_4BIT		(2)

	int now;
	int (*enable_wide)(struct mmc_card *card);
	int (*disable_wide)(struct mmc_card *card);
};

static int bus_width = 4;

static struct sdio_bus_type bus_info = {
	.now = -1,
};

int aerial_log_flag = LOGLVL_1 | LOGLVL_2 ;

int
aerial_print(int level, const char *format, ...)
{
	if (aerial_log_flag & level) {
		int ret;
		va_list args;
		va_start(args, format);
		ret = vprintk(format, args);
		va_end(args);
		return ret;
	}
	return 0;
}

int
aerial_dump(int level, void *buf, int len)
{
	unsigned char *ptr = buf;
	int i;

	if (aerial_log_flag & level) {
		for (i=0; i<len; i++) {
			printk("%02x ", ptr[i]);
			if (i%16 == 15)
				printk("\n");
		}
		if (i%16 != 0)
			printk("\n");
	}
	return 0;
}

static int
aerial_sdio_f0_readb(struct sdio_func *func, unsigned int addr,
		     int *err_ret)
{
	int __num = func->num;
	unsigned char val;
	func->num = 0;
	val = sdio_readb(func, addr, err_ret);
	func->num = __num;
	return val;
}

static void
aerial_sdio_f0_writeb(struct sdio_func *func, unsigned char b,
		      unsigned int addr, int *err_ret)
{
	int __num = func->num;
	func->num = 0;
	sdio_writeb(func, b, addr, err_ret);
	func->num = __num;
}

static int
sdio_card_reset(struct mmc_card *card)
{
	struct sdio_func *func = card->sdio_func[0];
	unsigned char val;
	int ret;

	val = aerial_sdio_f0_readb(func, SDIO_CCCR_ABORT, &ret);
	if (ret)
		return ret;
	val |= 0x08;
	aerial_sdio_f0_writeb(func, val, SDIO_CCCR_ABORT, &ret);
	if (ret != -ETIMEDOUT)
		return ret;

	return 0;
}

struct aerial_restart_data {
	struct mmc_host *host;
	struct timer_list *timer;
	int state;
};
static struct aerial_restart_data restart_data;
static struct timer_list restart_timer;

static void
aerial_restart_timer_func(unsigned long _data)
{
	struct aerial_restart_data *data = (struct aerial_restart_data *)_data;
	struct mmc_host *host = data->host;
	struct timer_list *timer = data->timer;

	mmc_detect_change(host, msecs_to_jiffies(0));

	if (data->state++ == 0)
		mod_timer(timer, jiffies + msecs_to_jiffies(100));
}

int
aerial_restart(struct aerial_private *priv)
{
	struct sdio_func *func = priv->func;
	struct mmc_card *card = func->card;
	struct mmc_host *host = card->host;

	if (atomic_read(&priv->state) == AERIAL_STATE_NO_FW)
		return 0;

	sdio_claim_host(func);
	sdio_card_reset(card);
	sdio_release_host(func);

	restart_data.host = host;
	restart_data.timer = &restart_timer;
	restart_data.state = 0;
	setup_timer(restart_data.timer, aerial_restart_timer_func,
		    (unsigned long)&restart_data);
	mod_timer(restart_data.timer, jiffies + msecs_to_jiffies(1000));

	return 0;
}

static int
aerial_enable_wide(struct mmc_card *card)
{
	if (bus_info.now != 4) {
		sdio_enable_wide(card);
		bus_info.now = 4;
	}
	return 0;
}

static int
aerial_disable_wide(struct mmc_card *card)
{
	if (bus_info.now != 1) {
		sdio_disable_wide(card);
		bus_info.now = 1;
	}
	return 0;
}

/* sdio-lock of this function must be taken. */
int
aerial_send_prepare(struct sdio_func *func, int size)
{
	unsigned char val;
	int retry = 100;
	int ret = -EIO;

	size = size & AERIAL_PACKET_LENGTH_MASK;
	if (size)
		size |= AERIAL_PACKET_LENGTH_VALID;

	/* Set block length (LSB) to TxPacketLength register */
	retry = 10;
	while (retry--) {
		sdio_writeb(func, size & 0xff,
			    AERIAL_TX_PACKET_LENGTH_LSB, &ret);
		if (!ret) {
			val = sdio_readb(func,
					 AERIAL_TX_PACKET_LENGTH_LSB,
					 &ret);
			if (val == (size & 0xff))
				break;
		}
	}
	if (ret) {
		aer_err("failed set TxPacketLength register(LSB)\n");
		return -EIO;
	}

	/* Set block length (MSB) to TxPacketLength register */
	retry = 10;
	while (retry--) {
		sdio_writeb(func, (size >> 8),
			    AERIAL_TX_PACKET_LENGTH_MSB, &ret);
		if (!ret) {
			val = sdio_readb(func,
					 AERIAL_TX_PACKET_LENGTH_MSB,
					 &ret);
			if (val == (size >> 8))
				break;
		}
	}
	if (ret) {
		aer_err("failed set TxPacketLength register(MSB)\n");
		return -EIO;
	}

	return 0;
}

static int
aerial_set_interrupt(struct sdio_func *func, int set)
{
	unsigned char val;
	int ret;

	sdio_writeb(func, set, AERIAL_INT_MASK, &ret);
	if (ret) {
		aer_err("failed set f1:int-mask\n");
		return -1;
	}
	val = sdio_readb(func, AERIAL_INT_MASK, &ret);
	if (ret || (val != set)) {
		aer_err("failed read f1:int-mask (0x%x)\n", val);
		return -1;
	}
	aer_debug("TX Transmission mode %s\n",
		  (val & AERIAL_INT_TXRDY) ? "SDINT HT" : "CIS POLLING");
	aer_debug("RX Transmission mode %s\n",
		  (val & AERIAL_INT_RXRCV) ? "SDINT HT" : "CIS POLLING");

	return 0;
}

int
aerial_flow_start(struct aerial_private *priv)
{
	struct sdio_func *func = priv->func;
	unsigned long start;
	unsigned char val;
	int ret;

	if (atomic_read(&priv->power_mngt) != AERIAL_PWR_MNGT_ACTIVE) {
		if (aerial_send_prepare(func, AERIAL_WRITE_WORD_SIZE)) {
			aer_err("failed send prepare(TX data waiting)\n");
			return -EINVAL;
		}
	}

	start = jiffies;
	for (;;) {
		val = aerial_sdio_f0_readb(func, AERIAL_TX_CISINFO_STATUS, &ret);
		if (ret) {
			aer_err("failed read : CIS INFO STATUS(TX)\n");
			return -EIO;
		}
		if (IS_TXWAIT(val)) {
			if (IS_TOGGLE(val, priv->tx_toggle)) {
				SET_TOGGLE(val, priv->tx_toggle);
				break;
			}
		}
		else if (IS_TXBUSY(val)) {
			priv->tx_toggle = TOGGLE_NONE;
		}

		if (priv->tx_cis_toggle_timeout &&
		    time_after(jiffies, start + priv->tx_cis_toggle_timeout)) {
			priv->tx_toggle = TOGGLE_NONE;
			aer_err("TxToggle doesn't adjust. "
				"(timeout = %dmsec, reg = 0x%x)\n",
				jiffies_to_msecs(priv->tx_cis_toggle_timeout),
				val);
			break;
		}
		udelay(priv->tx_cis_intval);
	}

	return 0;
}

int
aerial_flow_end(struct aerial_private *priv)
{
	struct sdio_func *func = priv->func;
	int ret;

	if (atomic_read(&priv->power_mngt) != AERIAL_PWR_MNGT_ACTIVE) {
		ret = aerial_send_prepare(func, 0x00);
		if (ret) {
			aer_err("failed send prepare(TX data complete)\n");
			return -EINVAL;
		}
	}

	return 0;
}

static void
aerial_rx_task(struct sdio_func *func)
{
	struct aerial_private *priv = sdio_get_drvdata(func);
	struct aerial_packet *pr = &priv->rxbuf;
	struct sk_buff *skb;
	int size;
	int ret;

	aerial_led_trigger_event(priv, active, LED_FULL);

	ret = sdio_readsb(func, (char *)&priv->rxbuf,
			  AERIAL_RX_BUFFER, AERIAL_WRITE_SIZE);
	if (ret) {
		aer_err("failed first block read (stat = 0x%x)\n",
			sdio_readb(func, AERIAL_INT_STATUS, &ret));
		priv->net_stats.rx_errors++;
		goto RX_END;
	}

	size = pr->h.len + 2;
	if (size > AERIAL_READ_SIZE) {
		aer_err("unknown size [%d]\n", size);
		priv->net_stats.rx_errors++;
		goto RX_END;
	}

	if (pr->h.type == HTYPE_CONFIG_RES) {
		if (pr->m.type == CTYPE_RESPONSE) {
			memcpy(&priv->wid_response,
			       (char *)&priv->rxbuf, size);
			complete(&priv->wid_complete);
			goto RX_END;
		}
		if (pr->m.type == CTYPE_INFO) {
			struct aerial_wid_frame *rf = to_wid_frame(&priv->rxbuf);
			if (rf->wid == WID_STATUS) {
				if (pr->message[7] == 0x0) {
					aer_debug("DISCONNECT event received.\n");
					aerial_led_trigger_event(priv, link, LED_OFF);
				} else {
					aer_debug("CONNECT event received.\n");
					aerial_led_trigger_event(priv, link, LED_FULL);
				}
			}
			else if (rf->wid == WID_DEVICE_READY) {
				if (atomic_read(&priv->state) ==
				    AERIAL_STATE_NO_FW)
					complete(&priv->firmware_ready);
			}
			goto RX_END;
		}
		goto RX_END;
	}

	/* Data IN packet */
	skb = dev_alloc_skb(size + NET_IP_ALIGN);
	if (!skb) {
		priv->net_stats.rx_dropped++;
		aer_err("failed skb memory allocate error\n");
		goto RX_END;
	}
	skb->dev = priv->netdev;
	skb_reserve(skb, NET_IP_ALIGN);
	skb_put(skb, size);
	size = size - sizeof(struct aerial_packet_header);
	skb_copy_to_linear_data(skb, pr->message, size);
	skb->protocol = eth_type_trans(skb, skb->dev);

	ret = netif_rx(skb);
	if (ret == NET_RX_DROP) {
		priv->net_stats.rx_dropped++;
		aer_debug("packet dropped\n");
		dev_kfree_skb_any(skb);
	}
	priv->net_stats.rx_packets++;
	priv->net_stats.rx_bytes += size;

RX_END:
	aerial_led_trigger_event(priv, active, LED_OFF);

	return;
}

static void
aerial_tx_task(struct sk_buff *skb, struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);
	struct aerial_packet txbuf;
	struct sdio_func *func = priv->func;
	unsigned long flags;
	int ret;

	if (atomic_read(&priv->state) != AERIAL_STATE_RUNNING) {
		aer_err("tx_task: driver state is not running!\n");
		dev_kfree_skb_any(skb);
		return;
	}

	txbuf.h.len = skb->len;
	txbuf.h.type = HTYPE_DATA_OUT;
	memcpy(txbuf.message, skb->data, skb->len);
	dev_kfree_skb_any(skb);

	dev->trans_start = jiffies;

	local_irq_save(flags);
	if (atomic_read(&priv->power_state) == AERIAL_PWR_STATE_SLEEP) {
		local_irq_restore(flags);
		return;
	}
	sdio_claim_host(func);
	local_irq_restore(flags);

	aerial_led_trigger_event(priv, active, LED_FULL);

	if (bus_info.enable_wide)
		bus_info.enable_wide(func->card);

	ret = aerial_flow_start(priv);
	if (ret)
		goto tx_end;

	ret = sdio_writesb(func, AERIAL_TX_BUFFER, &txbuf,
			   AERIAL_WRITE_SIZE);
	if (ret) {
		aer_err("failed send data\n");
		priv->net_stats.tx_errors++;
		goto tx_end;
	}

	ret = aerial_flow_end(priv);
	if (ret)
		goto tx_end;

	priv->net_stats.tx_packets++;
	priv->net_stats.tx_bytes += txbuf.h.len;

tx_end:
	if (bus_info.disable_wide)
		bus_info.disable_wide(func->card);
	sdio_release_host(func);

	aerial_led_trigger_event(priv, active, LED_OFF);

	return;
}

static void
aerial_irq(struct sdio_func *func)
{
	unsigned char status, val;
	int ret;

	if (bus_info.enable_wide)
		bus_info.enable_wide(func->card);

	/* Check Interrupt status */
	status = sdio_readb(func, AERIAL_INT_STATUS, &ret);
	if (ret) {
		aer_err("failed get interrupt status\n");
		return;
	}
	if (status & (AERIAL_INT_RXRCV | AERIAL_INT_TXRDY)) {
		sdio_writeb(func, status, AERIAL_INT_PENDING, &ret);
		if (ret)
			aer_err("failed clear interrupt status\n");
		for (;;) {
			val = sdio_readb(func, AERIAL_INT_STATUS, &ret);
			if (ret)
				aer_err("failed get interrupt status[2]\n");
			if ((status & val) == 0)
				break;
			aer_err("interrupt pending? (0x%x)\n", val);
		}
	}

	if (status & AERIAL_INT_RXRCV)
		aerial_rx_task(func);

	if (bus_info.disable_wide)
		bus_info.disable_wide(func->card);

	return;
}

static int
aerial_open(struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);

	if (atomic_read(&priv->state) == AERIAL_STATE_NO_FW) {
		aer_err("firmware was not loadding.\n");
		return -EIO;
	}

	atomic_set(&priv->state, AERIAL_STATE_RUNNING);
	netif_start_queue(dev);

	return 0;
}

static int
aerial_close(struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);

	atomic_set(&priv->state, AERIAL_STATE_READY);
	netif_stop_queue(dev);

	cancel_delayed_work(&priv->scan_task);
	cancel_delayed_work(&priv->link_task);

	return 0;
}

static int
aerial_hardstart(struct sk_buff *skb, struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);
	struct aerial_task_info *task = &priv->tx_task;
	struct tx_args *args;
	unsigned long flags;

	args = kzalloc(sizeof(struct tx_args), GFP_KERNEL);
	if (args == NULL) {
		dev_kfree_skb_any(skb);
		aer_err("failed tx_args memory allocation\n");
		return -ENOMEM;
	}

	args->dev = dev;
	args->skb = skb;

	spin_lock_irqsave(&task->lock, flags);
	list_add_tail(&args->entry, &task->tx_args_list);
	spin_unlock_irqrestore(&task->lock, flags);

	schedule_delayed_work(&task->queue, usecs_to_jiffies(0));

	return 0;
}

static int
aerial_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	return 0;
}

static struct net_device_stats *
aerial_getstats(struct net_device *dev)
{
	struct aerial_private *priv = netdev_priv(dev);

	return &priv->net_stats;
}

static void
aerial_do_task(struct work_struct *work)
{
	struct aerial_task_info *task = container_of(work,
						     struct aerial_task_info,
						     queue.work);
	struct aerial_private *priv = container_of(task,
						   struct aerial_private,
						   tx_task);

	complete(&priv->tx_thread_wait);
}

static int
aerial_tx_thread(void *_priv)
{
	struct aerial_private *priv = _priv;
	struct aerial_task_info *task = &priv->tx_task;
	struct tx_args *args;
	unsigned long flags;

	init_completion(&priv->tx_thread_wait);

	do {
		int timeout;

		timeout = wait_for_completion_interruptible_timeout(
						&priv->tx_thread_wait,
						msecs_to_jiffies(100));
		set_current_state(TASK_RUNNING);
		if (timeout > 0) {
			spin_lock_irqsave(&task->lock, flags);
			while (!list_empty(&task->tx_args_list)) {
				args = list_entry(task->tx_args_list.next,
						  struct tx_args, entry);
				list_del(&args->entry);
				spin_unlock_irqrestore(&task->lock, flags);

				aerial_tx_task(args->skb, args->dev);
				kfree(args);

				spin_lock_irqsave(&task->lock, flags);
			}
			spin_unlock_irqrestore(&task->lock, flags);
		}
	} while (!kthread_should_stop());

	return 0;
}

static int
aerial_task_init(struct aerial_private *priv)
{
	struct aerial_task_info *task = &priv->tx_task;

	memset(task, 0, sizeof(struct aerial_task_info));

	spin_lock_init(&task->lock);
	INIT_LIST_HEAD(&task->tx_args_list);
	INIT_DELAYED_WORK(&task->queue, aerial_do_task);

	priv->tx_thread = kthread_run(aerial_tx_thread, priv, "%s.tx/%s",
				      AERIAL_DRVINFO, priv->netdev->name);
	if (IS_ERR(priv->tx_thread)) {
		aer_err("Could not create thread.\n");
		return PTR_ERR(priv->tx_thread);
	}

	return 0;
}

static void
aerial_task_exit(struct aerial_private *priv)
{
	if (priv->tx_thread) {
		kthread_stop(priv->tx_thread);
		priv->tx_thread = NULL;
	}
}

static int
aerial_sdio_if_init(struct sdio_func *func)
{
	int ret;

	sdio_claim_host(func);

	ret = sdio_enable_func(func);
	if (ret) {
		aer_err("failed enable func\n");
		goto init_end;
	}

	ret = aerial_set_interrupt(func, AERIAL_INT_RXRCV);
	if (ret) {
		aer_err("failed interrupt settings\n");
		goto init_end;
	}

	sdio_claim_irq(func, aerial_irq);

	/* Set block length */
	ret = sdio_set_block_size(func, AERIAL_CHUNK_SIZE);
	if (ret) {
		aer_err("failed set block size\n");
		goto init_end;
	}

	if (bus_info.disable_wide)
		bus_info.disable_wide(func->card);

	ret = aerial_send_prepare(func, AERIAL_WRITE_WORD_SIZE);
	if (ret)
		aer_err("failed send prepare\n");

init_end:
	sdio_release_host(func);

	return ret;
}

static int
sdio_aerial_probe(struct sdio_func *func, const struct sdio_device_id *id)
{
	struct aerial_private *priv;
	struct net_device *dev;
	int ret;

	/* net device initialize */
	dev = alloc_netdev(sizeof(struct aerial_private),
			   "wlan%d", ether_setup);
	if (!dev) {
		aer_err("could not allocate device\n");
		return -ENOMEM;
	}

	dev->open = aerial_open;
	dev->stop = aerial_close;
	dev->hard_start_xmit = aerial_hardstart;
	dev->do_ioctl = aerial_ioctl;
	dev->get_stats = aerial_getstats;

	aerial_wireless_attach(dev);

	/* card specific capability */
	func->card->ext_caps = MMC_CARD_CAPS_FORCE_BLKXMIT;
	switch (bus_width) {
	case 1:
		bus_info.id = SDIO_FORCE_1BIT;
		break;
	default:
		aer_info("unknown bus-width. using default FORCE 4bit mode\n");
		/* FALL THROUGH */
	case 4:
		bus_info.id = SDIO_FORCE_4BIT;
		break;
	}

	bus_info.now = -1;
	switch (bus_info.id) {
	case SDIO_DYNAMIC_WIDTH:
		bus_info.enable_wide = aerial_enable_wide;
		bus_info.disable_wide = aerial_disable_wide;
		break;
	case SDIO_FORCE_1BIT:
		bus_info.enable_wide = NULL;
		bus_info.disable_wide = aerial_disable_wide;
		break;
	case SDIO_FORCE_4BIT:
		func->card->ext_caps |= MMC_CARD_CAPS_FORCE_CLK_KEEP;
		bus_info.enable_wide = aerial_enable_wide;
		bus_info.disable_wide = NULL;
		break;
	}

	/* Aerial initialize */
	priv = dev->priv;
	priv->netdev = dev;
	priv->func = func;
	priv->tx_cis_intval = AERIAL_TX_CISPOLL_INTVAL;
	priv->tx_cis_toggle_timeout =
		msecs_to_jiffies(AERIAL_TX_TOGGLE_TIMEOUT);
	priv->tx_toggle = TOGGLE_NONE;

	sdio_set_drvdata(func, priv);

	atomic_set(&priv->state, AERIAL_STATE_NO_FW);
	atomic_set(&priv->power_state, AERIAL_PWR_STATE_ACTIVE);
	atomic_set(&priv->power_mngt, AERIAL_PWR_MNGT_ACTIVE);

	sema_init(&priv->wid_req_lock, 1);
	init_completion(&priv->wid_complete);
	init_completion(&priv->firmware_ready);
	INIT_DELAYED_WORK(&priv->scan_task, NULL);
	INIT_DELAYED_WORK(&priv->link_task, NULL);

	priv->fw_image = kzalloc(FIRMWARE_SIZE_MAX, GFP_KERNEL);
	priv->fw_size  = 0;

	ret = aerial_sdio_if_init(func);
	if (ret)
		goto if_err;

	ret = register_netdev(dev);
	if (ret) {
		aer_err("failed register netdev\n");
		goto netdev_err;
	}

	ret = aerial_sysfs_init(priv);
	if (ret) {
		aer_err("failed register sysfs-entry\n");
		goto sysfs_err;
	}

	ret = aerial_task_init(priv);
	if (ret)
		goto task_err;

	aerial_led_trigger_register(priv, active);
	aerial_led_trigger_register(priv, link);
	aerial_led_trigger_event(priv, active, LED_OFF);
	aerial_led_trigger_event(priv, link, LED_OFF);

	printk(KERN_INFO "%s: registerd \"%s\" device as %s\n",
	       mmc_hostname(func->card->host), AERIAL_DRVINFO, dev->name);

	return 0;

task_err:
	aerial_sysfs_remove(priv);
sysfs_err:
	unregister_netdev(dev);
netdev_err:
if_err:
	free_netdev(dev);

	return ret;
}

static void
sdio_aerial_remove(struct sdio_func *func)
{
	struct aerial_private *priv = sdio_get_drvdata(func);

	aerial_led_trigger_unregister(priv, active);
	aerial_led_trigger_unregister(priv, link);

	aerial_task_exit(priv);

	sdio_claim_host(func);
	sdio_release_irq(func);
	sdio_disable_func(func);
	sdio_release_host(func);

	unregister_netdev(priv->netdev);
	aerial_sysfs_remove(priv);

	kfree(priv->fw_image);
	free_netdev(priv->netdev);

	return;
}

static const struct sdio_device_id sdio_aerial_ids[] = {
	{ SDIO_CLASS_WLAN, SDIO_VENDOR_ID_ROHM, SDIO_DEVICE_ID_ROHM_BU1806GU },
	{ /* end: all zeros */ },
};

MODULE_DEVICE_TABLE(sdio, sdio_aerial_ids);

static struct sdio_driver sdio_aerial_driver = {
	.probe = sdio_aerial_probe,
	.remove = sdio_aerial_remove,
	.name = "sdio_aerial",
	.id_table = sdio_aerial_ids,
};

static int __init
sdio_aerial_init(void)
{
	int ret;

	ret = sdio_register_driver(&sdio_aerial_driver);
	if (ret)
		aer_err("Registration failure.(%d)\n", ret);
	else
		aer_info("Version %s Load.\n", AERIAL_VERSION);

	return ret;
}

static void __exit
sdio_aerial_exit(void)
{
	sdio_unregister_driver(&sdio_aerial_driver);
	aer_info("Version %s unloaded.\n", AERIAL_VERSION);
}

module_init(sdio_aerial_init);
module_exit(sdio_aerial_exit);

module_param(bus_width, int, 0);

MODULE_AUTHOR("Nissin Systems Co.,Ltd.");
MODULE_LICENSE("GPL");
