/*
 * drivers/mmc/card/aerial/aerial_fw.c
 *
 *  Copyright (C) 2008 Nissin Systems Co.,Ltd.
 *  Copyright (C) 2008-2009 Atmark Techno, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 2008-04-02    Created by Nissin Systems Co.,Ltd.
 * 2009-06-19   Modified by Atmark Techno, Inc.
 */

#include <linux/module.h>
#include <linux/mmc/host.h>
#include <linux/mmc/card.h>
#include <linux/mmc/sdio_func.h>

#include "aerial_drv.h"
#include "aerial_wid.h"
#include "aerial_fw.h"

#define WID_DEFAULT_TIMEOUT (100)

#define to_download_addr(w) ((w)->val[3] << 24 | (w)->val[2] << 16 | \
			     (w)->val[1] <<  8 | (w)->val[0] <<  0)
#define to_download_size(w) ((w)->val[5] <<  8 | (w)->val[4] <<  0)

int
aerial_firmware_load(struct sdio_func *func)
{
	struct aerial_private *priv = sdio_get_drvdata(func);
	struct aerial_packet req;
	struct aerial_packet res;
	struct aerial_wid_frame *wframe = to_wid_frame(&req);
	int ret;
	struct aerial_firmware_info *info =
		(struct aerial_firmware_info *)priv->fw_image;
	unsigned char val;

	if (info->id != AERIAL_FW_ID)
		return -EINVAL;

	priv->tx_toggle = TOGGLE_0;
	memset(&req, 0, sizeof(struct aerial_packet));
	memset(&res, 0, sizeof(struct aerial_packet));

	val = 1;
	set_wid_write_request(&req, WID_RESET, &val, 1);

	ret = aerial_wid_request(func, &req, &res, WID_DEFAULT_TIMEOUT);
	if (ret) {
		aer_err("failed WID_RESET\n");
		return -EIO;
	}

	wframe = to_wid_frame(&res);
	while (wframe->wid == WID_DOWNLOAD_REQUEST) {
		struct aerial_wid_longframe *lframe = to_wid_longframe(&req);
		unsigned char sum;
		unsigned char *src;
		unsigned char *dst;
		unsigned long request_addr;
		unsigned short request_size;
		unsigned long loop;

		request_addr = to_download_addr(wframe);
		request_size = to_download_size(wframe);

		memset(&req, 0, sizeof(struct aerial_packet));

		/* wid */
		lframe->wid = WID_DOWNLOAD_DATA;
		lframe->len = request_size;

		sum = 0;
		src = ((unsigned char*)info) + request_addr;
		dst = lframe->val;
		for (loop = 0; loop < request_size; loop++, src++, dst++) {
			*dst = *src;
			sum += *dst;
		}
		*dst = sum;

		/* message */
		req.m.len = (aerial_packet_message_fixed_len
			     + 2 /* wid:(unsinged short) */
			     + 2 /* len:(unsigned short) */
			     + 1 /* sum:(unsigned char) */
			     + request_size);

		req.m.type = CTYPE_WRITE;

		/* header */
		req.h.len = req.m.len;
		req.h.type = HTYPE_CONFIG_REQ;

		memset(&res, 0, sizeof(struct aerial_packet));
		ret = aerial_wid_request(func, &req, &res, WID_DEFAULT_TIMEOUT);
		if (ret) {
			aer_err("failed WID_DOWNLOAD_DATA (err = %d)\n", ret);
			if (ret < 0)
				return ret;
			else
				return -EIO;
		}
	}

	if (wframe->wid != WID_START) {
		aer_err("unknown WID received\n");
		return -EINVAL;
	}

	memset(&req, 0, sizeof(struct aerial_packet));

	val = 1;
	set_wid_write_request(&req, WID_START, &val, 1);

	ret = aerial_wid_request(func, &req, NULL, WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed WID_START\n");

	return ret;
}

int
aerial_firmware_setup(struct aerial_private *priv)
{
	unsigned char mac[6] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
	unsigned long timeout;
	char version[128];
	int ret;

	if (atomic_read(&priv->state) != AERIAL_STATE_NO_FW) {
		aer_debug("Firmware is already loaded\n");
		return 0;
	}

	ret = aerial_firmware_load(priv->func);
	if (ret)
		return ret;

	priv->tx_toggle = TOGGLE_NONE;

	timeout = msecs_to_jiffies(500);
	timeout = wait_for_completion_timeout(&priv->firmware_ready, timeout);
	if (timeout <= 0)
		return -EIO;
	atomic_set(&priv->state, AERIAL_STATE_READY);
	aerial_get_macaddr(priv, mac, 6);
	memcpy(priv->netdev->dev_addr, mac, 6);

	aer_info("MAC is %02x:%02x:%02x:%02x:%02x:%02x\n",
	       mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);

	aerial_get_firmware_version(priv, version, 128);
	if (version[0] == 'T')
		priv->fw_type = AERIAL_FWTYPE_TEST;
	else
		priv->fw_type = AERIAL_FWTYPE_STA;

	return 0;
}
