/*
 * drivers/mmc/card/aerial/aerial_wid.c
 *
 *  Copyright (C) 2008 Atmark Techno, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 2008-12-11    Created by Atmark Techno, Inc.
 * 2009-06-19   Modified by Atmark Techno, Inc.
 */

#include <linux/netdevice.h>
#include <linux/delay.h>
#include <linux/mmc/host.h>
#include <linux/mmc/card.h>
#include <linux/mmc/sdio_func.h>
#include <net/iw_handler.h>
#include "aerial_drv.h"
#include "aerial_wid.h"

#define WID_DEFAULT_TIMEOUT	(100)
#define WID_GENKEY_TIMEOUT	(20000)

#define NOFW_EXIT(p)							\
({									\
	if (atomic_read(&((p)->state)) == AERIAL_STATE_NO_FW)		\
		return -EINVAL;						\
})
#define SLEEP_EXIT(p)							\
({									\
	if (atomic_read(&((p)->power_state)) == AERIAL_PWR_STATE_SLEEP)	\
		return -EBUSY;						\
})

static int
aerial_set_common_value(struct aerial_private *priv, unsigned short wid,
			void *buf, int size, unsigned int timeout_ms)
{
	struct sdio_func *func = priv->func;
	struct aerial_packet *req, _req;
	struct aerial_packet *res, _res;
	struct aerial_wid_frame *wframe = to_wid_frame(&_res);
	int ret;

	req = &_req;
	res = &_res;
	memset(req, 0, sizeof(struct aerial_packet));
	memset(res, 0, sizeof(struct aerial_packet));

	if (timeout_ms == -1) {
		res = NULL;
		timeout_ms = 1;
	}

	set_wid_write_request(req, wid, buf, size);
	ret = aerial_wid_request(func, req, res, timeout_ms);
	if (ret)
		return ret;

	if (res) {
		if (!response_ok(wframe))
			return -EIO;
	}

	return 0;
}

static int
aerial_get_common_value(struct aerial_private *priv, unsigned short wid,
			void *buf, int size, unsigned int timeout_ms)
{
	struct sdio_func *func = priv->func;
	struct aerial_packet req;
	struct aerial_packet res;
	struct aerial_wid_frame *wframe = to_wid_frame(&res);
	int ret;

	memset(&req, 0, sizeof(struct aerial_packet));
	memset(&res, 0, sizeof(struct aerial_packet));

	set_wid_query_request(&req, wid);
	ret = aerial_wid_request(func, &req, &res, timeout_ms);
	if (ret)
		return ret;

	if (size < wframe->len)
		return -EINVAL;

	memcpy(buf, wframe->val, wframe->len);

	return 0;
}

static int
is_wpa_mode(struct aerial_private *priv)
{
	unsigned char mode;
	int ret;

	ret = aerial_get_cryptmode(priv, &mode);
	if (ret)
		return -EIO;

	switch (mode) {
	case AERIAL_CRYPT_WPA_CCMP:
	case AERIAL_CRYPT_WPA_TKIP:
	case AERIAL_CRYPT_WPA2_CCMP:
	case AERIAL_CRYPT_WPA2_TKIP:
		return 1;
	default:
		return 0;
	}
}

int
aerial_set_bsstype(struct aerial_private *priv, unsigned char type)
{
	int range = 1;
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (priv->fw_type == AERIAL_FWTYPE_TEST)
		range = 6;
	if ((type != AERIAL_BSS_WLANOFF) && (range < type))
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_BSS_TYPE, &type, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set BSS type\n");

	return ret;
}

int
aerial_get_bsstype(struct aerial_private *priv, unsigned char *type)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_BSS_TYPE, type, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get BSS type\n");

	return ret;
}

int
aerial_set_ssid(struct aerial_private *priv, char *ssid, int size)
{
	int timeout;
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (size < 1 || 32 < size)
		return -EINVAL;

	ret = is_wpa_mode(priv);
	switch (ret) {
	case 1:
		timeout = WID_GENKEY_TIMEOUT;
		break;
	case 0:
		timeout = WID_DEFAULT_TIMEOUT;
		break;
	default:
		return ret;
	}

	ret = aerial_set_common_value(priv, WID_SSID, ssid, size,
				      timeout);
	if (ret)
		aer_err("failed set SSID\n");

	return ret;
}

int
aerial_get_ssid(struct aerial_private *priv, char *ssid, int size)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (size < 32)
		return -EINVAL;

	ret = aerial_get_common_value(priv, WID_SSID, ssid, size,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get SSID\n");

	return ret;
}

int
aerial_get_bssid(struct aerial_private *priv, unsigned char *bssid, int size)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (size < 6)
		return -EINVAL;

	ret = aerial_get_common_value(priv, WID_BSSID, bssid, size,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get BSSID\n");

	return ret;
}

int
aerial_set_channel(struct aerial_private *priv, unsigned char ch)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (ch == 0 || 13 < ch)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_CURRENT_CHANNEL, &ch, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set channel\n");

	return ret;
}

int
aerial_get_channel(struct aerial_private *priv, unsigned char *ch)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_CURRENT_CHANNEL, ch, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get channel\n");

	return ret;
}

int
aerial_set_macaddr(struct aerial_private *priv, unsigned char *mac, int size)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (priv->fw_type != AERIAL_FWTYPE_TEST)
		return -EOPNOTSUPP;
	if (size != 6)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_MAC_ADDR, mac, 6,
				      1000);
	if (ret)
		aer_err("failed set MAC address\n");

	return ret;
}

int
aerial_get_macaddr(struct aerial_private *priv, unsigned char *mac, int size)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (size < 6)
		return -EINVAL;

	ret = aerial_get_common_value(priv, WID_MAC_ADDR, mac, 6,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get MAC address\n");

	return ret;
}

int
aerial_set_rate(struct aerial_private *priv, unsigned char rate)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	switch (rate) {
	case  0: case  1: case  2: case  5: case  6:
	case  9: case 11: case 12: case 18: case 24:
	case 36: case 48: case 54:
		break;
	default:
		return -EINVAL;
	}

	ret = aerial_set_common_value(priv, WID_CURRENT_TX_RATE, &rate, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set Tx rate\n");

	return ret;
}

int
aerial_get_rate(struct aerial_private *priv, unsigned char *rate)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_CURRENT_TX_RATE, rate, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get rate\n");

	return ret;
}

int
aerial_set_cryptmode(struct aerial_private *priv, unsigned char mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_set_common_value(priv, WID_802_11I_MODE, &mode, 1,
				      WID_GENKEY_TIMEOUT);
	if (ret)
		aer_err("failed set crypt mode\n");

	return ret;
}

int
aerial_get_cryptmode(struct aerial_private *priv, unsigned char *mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_802_11I_MODE, mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get crypt mode\n");

	return ret;
}

int
aerial_set_wepkey(struct aerial_private *priv, int index, char *key, int len)
{
	unsigned short wid_wep_key[] = { WID_WEP_KEY_VALUE0 };
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (len != 10 && len != 26)
		return -EINVAL;
	if (index != 0)
		return -EINVAL;

	ret = aerial_set_common_value(priv, wid_wep_key[index], key, len,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set WEP key\n");

	return ret;
}

int
aerial_get_wepkey(struct aerial_private *priv, int index, char *key, int len)
{
	unsigned short wid_wep_key[] = { WID_WEP_KEY_VALUE0 };
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (len < 26)
		return -EINVAL;
	if (index != 0)
		return -EINVAL;

	ret = aerial_get_common_value(priv, wid_wep_key[index], key, len,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get WEP key\n");

	return ret;
}

int
aerial_set_powerman(struct aerial_private *priv, unsigned char mode)
{
	struct sdio_func *func = priv->func;
	unsigned char current_mode;
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (1 < mode)
		return -EINVAL;

	ret = aerial_get_powerman(priv, &current_mode);
	if (ret)
		return ret;

	if (mode == AERIAL_PWR_MNGT_ACTIVE) {
		atomic_set(&priv->power_mngt, AERIAL_PWR_MNGT_ACTIVE);
		sdio_claim_host(func);
		ret = aerial_send_prepare(func, AERIAL_WRITE_WORD_SIZE);
		sdio_release_host(func);
		if (ret) {
			atomic_set(&priv->power_mngt, current_mode);
			return ret;
		}
	}

	ret = aerial_set_common_value(priv, WID_POWER_MANAGEMENT, &mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret) {
		atomic_set(&priv->power_mngt, current_mode);
		aer_err("failed set Power management\n");
		return ret;
	}

	if (mode == AERIAL_PWR_MNGT_SAVE) {
		sdio_claim_host(func);
		ret = aerial_send_prepare(func, 0x00);
		sdio_release_host(func);
		if (ret)
			return ret;
		atomic_set(&priv->power_mngt, AERIAL_PWR_MNGT_SAVE);
	}

	return ret;
}

int
aerial_get_powerman(struct aerial_private *priv, unsigned char *mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_POWER_MANAGEMENT, mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get Power management\n");

	return ret;
}

int
aerial_set_scan_type(struct aerial_private *priv, unsigned char type)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_set_common_value(priv, WID_SCAN_TYPE, &type, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set scan type\n");

	return ret;
}

int
aerial_get_scan_type(struct aerial_private *priv, unsigned char *type)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_SCAN_TYPE, type, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get scan type\n");

	return ret;
}

int
aerial_set_sitesv(struct aerial_private *priv, unsigned char mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (2 < mode)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_SITE_SURVEY, &mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set site survey mode\n");

	return ret;
}

int
aerial_get_sitesv(struct aerial_private *priv, unsigned char *mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_SITE_SURVEY, mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get site survey mode\n");

	return ret;
}

int
aerial_set_psctl(struct aerial_private *priv, unsigned char mode)
{
	struct sdio_func *func = priv->func;
	struct mmc_card *card = func->card;
	struct mmc_host *host = card->host;
	int ret;

	NOFW_EXIT(priv);

	switch (mode) {
	case 2:
		if (atomic_read(&priv->power_state) == AERIAL_PWR_STATE_SLEEP)
			return 0;

		break;
	case 0:
		if (atomic_read(&priv->power_state) == AERIAL_PWR_STATE_ACTIVE)
			return 0;
		atomic_set(&priv->power_state, AERIAL_PWR_STATE_ACTIVE);

		if (host->ops->clk_start)
			host->ops->clk_start(host);

		sdio_release_host(func);
		break;
	default:
		return -EINVAL;
	}

	ret = aerial_set_common_value(priv, WID_POWER_SAVE, &mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set power save\n");

	switch (mode) {
	case 2:
		sdio_claim_host(func);
		atomic_set(&priv->power_state, AERIAL_PWR_STATE_SLEEP);
		if (host->ops->clk_stop)
			host->ops->clk_stop(host);
		break;
	default:
		break;
	}

	return ret;
}

int
aerial_set_startscan(struct aerial_private *priv, unsigned char mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (1 < mode)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_START_SCAN_REQ, &mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed site survey start request\n");

	return ret;
}

int
aerial_get_startscan(struct aerial_private *priv, unsigned char *mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_START_SCAN_REQ, mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get scan status\n");

	return ret;
}

int
aerial_set_beaconint(struct aerial_private *priv, unsigned short val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_set_common_value(priv, WID_BEACON_INTERVAL, &val, 2,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set beacon interval\n");

	return ret;
}

int
aerial_get_beaconint(struct aerial_private *priv, unsigned short *val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_BEACON_INTERVAL, val, 2,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get beacon interval\n");

	return ret;
}

int
aerial_set_lisnintvl(struct aerial_private *priv, unsigned char val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (val == 0 || 50 < val)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_LISTEN_INTERVAL, &val, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set listen interval\n");

	return ret;
}

int
aerial_get_lisnintvl(struct aerial_private *priv, unsigned char *val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_LISTEN_INTERVAL, val, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get listen interval\n");

	return ret;
}

int
aerial_set_passps(struct aerial_private *priv, char *passps, int size)
{
	int timeout;
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (size < 8 || 63 < size)
		return -EINVAL;

	ret = is_wpa_mode(priv);
	switch (ret) {
	case 1:
		timeout = WID_GENKEY_TIMEOUT;
		break;
	case 0:
		timeout = WID_DEFAULT_TIMEOUT;
		break;
	default:
		return ret;
	}

	ret = aerial_set_common_value(priv, WID_802_11I_PSK, passps, size,
				      timeout);
	if (ret)
		aer_err("failed set Pre-Shared Key passphrase\n");

	return ret;
}

int
aerial_set_antenna(struct aerial_private *priv, unsigned char mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (2 < mode)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_ANTENNA_SELECTION, &mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set Anttenna selection\n");

	return ret;
}

int
aerial_get_antenna(struct aerial_private *priv, unsigned char *mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_ANTENNA_SELECTION, mode, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get Antenna selection\n");

	return ret;
}

int
aerial_set_join(struct aerial_private *priv, unsigned char index)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (9 < index)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_JOIN_REQ, &index, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed network join\n");

	return ret;
}

int
aerial_get_firmware_version(struct aerial_private *priv, char *buf, int size)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (size < 128)
		return -EINVAL;

	memset(buf, 0, size);
	ret = aerial_get_common_value(priv, WID_FIRMWARE_VERSION, buf, size,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get firmware version\n");

	return ret;
}

int
aerial_get_site_survey_results(struct aerial_private *priv,
			       struct aerial_survey_res *buf, int size)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_SITE_SURVEY_RESULTS,
				      buf, size, WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get site survey results[2]\n");

	return ret;
}

int
aerial_get_rssi(struct aerial_private *priv, unsigned char *rssi)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_RSSI, rssi, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get rssi value\n");

	return ret;
}

int
aerial_get_current_mac_status(struct aerial_private *priv, unsigned char *val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_CURRENT_MAC_STATUS, val, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get current mac status\n");

	return ret;
}

int
aerial_get_serial_number(struct aerial_private *priv, char *buf, int size)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (size < 32)
		return -EINVAL;

	ret = aerial_get_common_value(priv, WID_SERIAL_NUMBER, buf, size,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get serial number\n");

	return ret;
}

int
aerial_set_scan_filter(struct aerial_private *priv, unsigned char val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_set_common_value(priv, WID_SCAN_FILTER, &val, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set scan filter\n");

	return ret;
}

int
aerial_get_scan_filter(struct aerial_private *priv, unsigned char *val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_SCAN_FILTER, val, 1,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get scan filter\n");

	return ret;
}

int
aerial_set_site_survey_scan_time(struct aerial_private *priv,
				 unsigned short val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (val < 10 || 60000 < val)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_SITE_SURVEY_SCAN_TIME, &val, 2,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set site survey scan time\n");

	return ret;
}

int
aerial_get_site_survey_scan_time(struct aerial_private *priv,
				 unsigned short *val)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_SITE_SURVEY_SCAN_TIME, val, 2,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get site survey scan time\n");

	return ret;
}

int
aerial_set_pmk(struct aerial_private *priv, char *key, int size)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (size != 32)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_802_11I_PMK, key, size,
				      WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed set Pairwise Master Key\n");

	return ret;
}

int
aerial_set_key_generation(struct aerial_private *priv, unsigned char mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	if (1 < mode)
		return -EINVAL;

	ret = aerial_set_common_value(priv, WID_802_11I_KEY_AUTO_GENERATION,
				      &mode, 1, WID_GENKEY_TIMEOUT);
	if (ret)
		aer_err("failed set key generation mode\n");

	return ret;
}

int
aerial_get_key_generation(struct aerial_private *priv, unsigned char *mode)
{
	int ret;

	NOFW_EXIT(priv);
	SLEEP_EXIT(priv);

	ret = aerial_get_common_value(priv, WID_802_11I_KEY_AUTO_GENERATION,
				      mode, 1, WID_DEFAULT_TIMEOUT);
	if (ret)
		aer_err("failed get key generation mode\n");

	return ret;
}

void
set_wid_query_request(struct aerial_packet *req,
		      unsigned short wid)
{
	struct aerial_wid_frame *wframe = to_wid_frame(req);

	/* WID part */
	wframe->wid = wid;

	/* MESSAGE part */
	req->m.type = CTYPE_QUERY;
	req->m.seqno = 0;
	req->m.len = (aerial_packet_message_fixed_len
		      + 2 /* wid:(unsinged short) */);

	/* HEADER part */
	req->h.type = HTYPE_CONFIG_REQ;
	req->h.len = req->m.len;
	req->h.len = req->m.len;
}

void
set_wid_write_request(struct aerial_packet *req,
		      unsigned short wid,
		      void *val,
		      unsigned char size)
{
	struct aerial_wid_frame *wframe = to_wid_frame(req);

	/* WID part */
	wframe->wid = wid;
	wframe->len = size;
	memcpy(wframe->val, val, size);

	/* MESSAGE part */
	req->m.type = CTYPE_WRITE;
	req->m.seqno = 0;
	req->m.len = (aerial_packet_message_fixed_len
		      + 2 /* wid:(unsinged short) */
		      + 1 /* len:(unsigned char) */
		      + size);

	/* HEADER part */
	req->h.type = HTYPE_CONFIG_REQ;
	req->h.len = req->m.len;
	req->h.len = req->m.len;
}

int
aerial_wid_request(struct sdio_func *func,
		   struct aerial_packet *req, struct aerial_packet *res,
		   unsigned int timeout_ms)
{
	struct aerial_private *priv = sdio_get_drvdata(func);
	static unsigned char seq = 0;
	long timeout;
	int ret;
	int retry = 3;

	ret = down_interruptible(&priv->wid_req_lock);
	if (ret)
		return -ERESTARTSYS;

	while (retry--) {

		memset(&priv->wid_response, 0, sizeof(struct aerial_packet));

		req->m.seqno = ++seq;

		sdio_claim_host(func);

		ret = aerial_flow_start(priv);
		if (ret) {
			sdio_release_host(func);
			goto cmd_end;
		}

		ret = sdio_writesb(func, AERIAL_TX_BUFFER,
				   req, AERIAL_WRITE_SIZE);
		if (ret) {
			aer_err("failed send wid request\n");
			sdio_release_host(func);
			goto cmd_end;
		}

		ret = aerial_flow_end(priv);

		sdio_release_host(func);

		if (ret)
			goto cmd_end;

		if (!res)
			goto cmd_end;

		timeout = msecs_to_jiffies(timeout_ms);
		while (timeout) {
			timeout = wait_for_completion_timeout(&priv->wid_complete,
							      timeout);
			if (timeout > 0) {
				if (atomic_read(&priv->state) != AERIAL_STATE_NO_FW) {
					if (priv->wid_response.m.seqno == seq)
						break;
				} else {
					struct aerial_wid_frame *rf = to_wid_frame(&priv->wid_response);
					if (rf->wid == WID_STATUS) {
						ret = (int)rf->val[0];
						memset(&priv->wid_response, 0x00, sizeof(struct aerial_packet));
						goto cmd_end;
					}
					break;
				}
			}
		}
		if (timeout)
			break;
	}
	if (retry < 0) {
		aer_err("wait wid response timed out\n");
		ret = -ETIMEDOUT;
		goto cmd_end;
	}
	memcpy(res, &priv->wid_response, priv->wid_response.h.len + 2);

cmd_end:
	up(&priv->wid_req_lock);

	return ret;
}
